

local config        = require "config"
local Signal        = require "source.library.signal"
local Actor         = require "source.actor.Actor"
local assets        = require "source.Assets"
local StateManager  = require "source.system.StateManager"
local Passenger     = Actor:extend()



--------------------------------------------------------------------------------------------------------- --
-- Variables --
--------------------------------------------------------------------------------------------------------- --

local default_config = {
  xvel            = 12.5,
  yvel            = 20,
  gravity         = 90,
  money           = 25,
  moneyDecayTime  = 1.5,
}
  
  

--------------------------------------------------------------------------------------------------------- --
-- States --
--------------------------------------------------------------------------------------------------------- --

local spawnState = function(self)
  local rnd = love.math.random(1,3)
  if rnd < 3 then self.sound.spawn[rnd]:play() end
  local update = function(self, dt)
    if self.y > self.currentStation.y-self.height then
      self:move(0, -self.yvel*dt)
    else
      self:setPosition(self.x, self.currentStation.y-self.height)
      if self.currentStation.taxi then
        self.taxi = self.currentStation.taxi
        self.stateManager:switch("Move")
      else self.stateManager:switch("Wait") end
    end
  end
  return update
end


local waitState = function(self)
  
  self:setAnimation("Wave")
  local update = function(self, dt)
    self:processCollisions(self.x, self.y)
    if self.money == 0 then self.stateManager:switch("Exit")
    elseif self.currentStation.taxi and not self.currentStation.taxi.occupied then
      self.taxi = self.currentStation.taxi
      self.stateManager:switch("Move")
    end
  end
  return update
end


local fixState = function(self, other)
  
  local update = function(self, dt)
    if other.x+other.width/2 < self.x+self.width/2 then
      self:setAnimation("MoveRight")
      self:move(self.xvel*dt, 0)
      self.x = math.min(self.x, self.currentStation.x+self.currentStation.width-self.width)
      if self.x > other.x+other.width then
        self.stateManager:switch("Wait")
      end
    else
      self:setAnimation("MoveLeft")
      self:move(-self.xvel*dt, 0)
      self.x = math.max(self.x, self.currentStation.x)
      if self.x+self.width < other.x then
        self.stateManager:switch("Wait")
      end
    end
  end
  return update
end


local moveState = function(self)
  
  local update = function(self, dt)
    -- Move to taxi
    if self.taxi.onGround and not self.taxi.occupied then 
      if self.taxi.x+self.taxi.width/2 > self.x+self.width/2 then
        self:setAnimation("MoveRight")
        self:move(self.xvel*dt, 0)
      else
        self:move(-self.xvel*dt, 0)
        self:setAnimation("MoveLeft")
      end
      if math.floor(self.taxi.x+self.taxi.width/2) == math.floor(self.x+self.width/2) then
        self.sound.station[self.destination.stationID]:play()
        self.stateManager:switch("Travel")
        Signal.emit("PassengerEntered", self.taxi)
      end
    else
      self.taxi = nil
      self.stateManager:switch("Wait")
    end
  end
  return update
end


local travelState = function(self)
  
  self:setPosition(-self.width, -self.height)
  local update = function(self, dt)
    if self.money == 0 and self.taxi.onGround then 
      Signal.emit("PassengerDelivered", self.taxi)
      self.stateManager:switch("Exit")
    end
  end
  return update
end


local exitState = function(self)
  
  self.collidable = false
  self:setPriority(3)
  if self.taxi then
    self:setPosition(self.taxi.x+self.taxi.width/2-self.width/2, self.currentStation.y-self.height)
    self.taxi = nil
  end
  local update = function(self, dt)
    if self.x < self.currentStation.x+self.currentStation.width-self.width then
      self:setAnimation("MoveRight")
      self:move(self.xvel*dt, 0)
    else
      self:setAnimation("Wave")
      self.currentAnimation:gotoFrame(1)
      self.currentAnimation:pause()
      self:move(0, self.yvel*dt)
      if self.y >= self.currentStation.y then self:destroy() end
    end
  end
  return update
end



--------------------------------------------------------------------------------------------------------- --
-- Local functions --
--------------------------------------------------------------------------------------------------------- --

local registerStates = function(self)
  
  self.stateManager = StateManager(self)
  self.stateManager:add("Spawn", spawnState)
  self.stateManager:add("Wait", waitState)
  self.stateManager:add("Fix", fixState)
  self.stateManager:add("Move", moveState)
  self.stateManager:add("Travel", travelState)
  self.stateManager:add("Exit", exitState)
  self.stateManager:switch("Spawn")
end


local registerSignals = function(self)
  
  -- Signals
  local taxiLandingEvent = function(taxi, station)
    if self.taxi then self.currentStation = station end
    if self.stateManager:getCurrentState() == "Exit" then return end
    if self.stateManager:getCurrentState() == "Wait" then
      if self.money == 0 then self.stateManager:switch("Exit") end
    end
    if self.stateManager:getCurrentState() == "Travel" then
      if self.money == 0 then
        Signal.emit("PassengerDelivered", taxi)
        self.stateManager:switch("Exit")
      elseif station == self.destination then
        self.currentStation = station
        local rnd = love.math.random(1,3)
        if rnd < 3 and self.money > 10 then self.sound.thanks[rnd]:play() end
        Signal.emit("PassengerDelivered", taxi, self)
        self.stateManager:switch("Exit")
      end
    end
  end
  local taxiExplodedEvent = function(taxi)
    if self.taxi and self.taxi == taxi then
      self.sound.noo:play()
      self:destroy()
    end
  end
  Signal.register("TaxiLanded", taxiLandingEvent)
  Signal.register("TaxiExploded", taxiExplodedEvent)
end


local loadAssets = function(self)
  
  local rnd = love.math.random(1, 2)
  if rnd == 1 then self:addAnimation("Wave", assets.passenger["WaveLeft"])
  else self:addAnimation("Wave", assets.passenger["WaveRight"]) end
  self:addAnimation("MoveLeft", assets.passenger["MoveLeft"])
  self:addAnimation("MoveRight", assets.passenger["MoveRight"])
  self:setAnimation("Wave")
  self.currentAnimation:gotoFrame(love.math.random(1, #self.currentAnimation.frames))
  self.sound = {
    spawn = {
      assets.getSound("whistle"),
      assets.getSound("heyTaxi"),
    },
    noo     = assets.getSound("noo"),
    station     = {
      assets.getSound("station1"),
      assets.getSound("station2"),
      assets.getSound("station3"),
      assets.getSound("station4"),
      assets.getSound("station5"),
      assets.getSound("station6"),
      assets.getSound("station7"),
      assets.getSound("station8"),
      assets.getSound("station9"),
      assets.getSound("station10"),
    },
    thanks  = {
      assets.getSound("thanks"),
      assets.getSound("thankYou") },
    }
end


local createMoneyTimer = function(self)
  local Level = require "source.system.Level"
  local funct = function() if self.money > 0 then self.money = self.money-1 end end
  Level.timer:after(2, function() Level.timer:every(self.moneyDecayTime, funct) end)
end


local drawPassenger = function(self)
  local x, y = self.currentStation.x*config._scale, (self.currentStation.y-self.height)*config._scale
  local width, height = self.currentStation.width*config._scale, (self.height+.5)*config._scale
  love.graphics.setScissor(x, y, width, height)
  self:drawAnimation(assets.spritesheet, self.x, self.y+1)
  love.graphics.setScissor()
end



--------------------------------------------------------------------------------------------------------- --
-- Passenger Class --
--------------------------------------------------------------------------------------------------------- --

function Passenger:new(startStation, endStation)
  Passenger.super.new(self)
  self:setType("Passenger")
  for k,v in pairs(default_config) do self[k] = v end
  self.startStation = startStation
  self.destination = endStation
  self.currentStation = startStation
  self.taxi = nil
  self:setPosition(love.math.random(startStation.x+1, startStation.x+startStation.width-self.width-1), startStation.y)
  self:setPriority(1)
  self:setDimensions(10, 15)
  loadAssets(self)
  registerStates(self)
  registerSignals(self)
  createMoneyTimer(self)
end


function Passenger:filter(other)
  if other.type == "Passenger" then return "cross" end
end


function Passenger:resolveCollision(other)
  if other.type == "Passenger" and self.stateManager:getCurrentState() == "Wait" then
    self.stateManager:switch("Fix", other)
  end
end


function Passenger:update(dt)
  self.stateManager:update(dt)
  self:updateAnimation(dt)
end


function Passenger:draw()
  drawPassenger(self)
  --love.graphics.print("STATE: "..self.stateManager:getCurrentState(), self.x, self.y-20)
  --love.graphics.printf(self.money, self.x, self.y-10, self.width, "center")
end

return Passenger

