

local config      = require "config"
local Signal      = require "source.library.signal"
local Actor       = require "source.actor.Actor"
local assets      = require "source.Assets"
local Taxi  = Actor:extend()



--------------------------------------------------------------------------------------------------------- --
-- Variables --
--------------------------------------------------------------------------------------------------------- --

local default_config = {
  ox                  = 3,
  oy                  = 3,
  xvel                = 0,
  yvel                = 0,
  xacc                = 220,
  yacc                = 240,
  maxVelocity         = 200,
  maxLandingVelocity  = 100,
  gravity             = 90,
  fuel                = 95,
  fuelConsumption     = 5,
  fuelCost            = 10,
  refuelrate          = 8.5,
  money               = 0,
  taxiHeight          = 15,
  occupied            = false,
  onGround            = false,
  gearDown            = false,
}



--------------------------------------------------------------------------------------------------------- --
-- Local Functions --
--------------------------------------------------------------------------------------------------------- --

local resolveLandingCollision = function(self, station)
  -- landing failed
  if self.y + self.height > station.y or -- Taxi too far off on y axis
  self.x < station.x-3 or self.x+self.width > station.x+station.width+3 -- Taxi too far off on x axis
  or self.yvel > self.maxLandingVelocity or not self.gearDown then -- Taxi to fast or landing gear not lowered
    self:destroy()
    return
  end
    
  -- Landing a bit to fast causes the taxi to bounce
  if self.yvel > self.maxLandingVelocity/2.25 then self.yvel = -self.yvel/2 return end
    
  -- Landing successful
  self.xvel = 0
  self.yvel = 0
  Signal.emit("TaxiLanded", self, station)
  self.onGround = true
end



local updateBoundaryBehaviour = function(self, dt)
  if self.x > config._resolutionX then self:setPosition(-self.width, self.y)
  elseif self.x+self.width < 0 then self:setPosition(config._resolutionX, self.y) end
  if self.y > config._resolutionY then self:setPosition(self.x, -self.height)
  elseif self.y+self.height < 0 then self:setPosition(self.x, config._resolutionY) end
end



local toggleGear = function(self)
  self.gearDown = not self.gearDown
  self.onGround = false
  if self.gearDown then self:setDimensions(self.width, self.taxiHeight-self.oy*2+3)
  else self:setDimensions(self.width, self.taxiHeight-self.oy*2) end
end



local updateKeyInput = function(self, dt)
  local keyDown = love.keyboard.isDown
  local left, right, up, down = keyDown("left"), keyDown("right"), keyDown("up"), keyDown("down")
  if up and self.yvel > -self.maxVelocity then
    self.yvel = self.yvel-self.yacc*dt
  end
  if not self.onGround then
    if down and self.yvel < self.maxVelocity then self.yvel = self.yvel+self.yacc*dt end
    if not self.gearDown then
      if left and self.xvel > -self.maxVelocity then
        self.xvel = self.xvel-self.xacc*dt
        self.flipped = false
      end
      if right and self.xvel < self.maxVelocity then
        self.xvel = self.xvel+self.xacc*dt
        self.flipped = true
      end
    end
  end
end



local applyMovement = function(self, dt)
  if not self.onGround and self.yvel < self.maxVelocity then self.yvel = self.yvel+self.gravity*dt end -- gravity
  self:move(self.xvel*dt, self.yvel*dt)
end



local updateFuel = function(self, dt)
  local keyDown = love.keyboard.isDown
  local left, right, up, down = keyDown("left"), keyDown("right"), keyDown("up"), keyDown("down")
  if (up or down or left or right) and not self.onGround then
    self.refueling = false
    self.fuel = self.fuel-self.fuelConsumption*dt
  end
  if self.fuel <= 30 then self.sound.noFuel:play()
  else self.sound.noFuel:stop() end
  if self.refueling and self.money > 0 and self.fuel < 100 then
    self.sound.refuel:play()
    self.fuel = math.min(self.fuel+self.refuelrate*dt, 100)
    self:removeMoney(self.fuelCost*dt)
  end
end



local drawTaxi = function(self)
  love.graphics.setScissor((self.x-self.ox)*config._scale,(self.y-self.oy)*config._scale, (self.width+self.ox*2)*config._scale, (self.height+self.oy+1)*config._scale)
  if self.flipped then self:drawSprite(assets.spritesheet, self.sprite, self.x, self.y, 0, -1, 1, self.width+self.ox, self.oy)
  else self:drawSprite(assets.spritesheet, self.sprite, self.x, self.y, 0, 1, 1, self.ox, self.oy) end
  love.graphics.setScissor()
end



local drawDebug = function(self)
  love.graphics.print("Money: "..math.floor(self.money), 10, 40)
  if self.fuel <= 30 then love.graphics.setColor(255, 0, 0) end
  love.graphics.print("Fuel: "..math.floor(self.fuel), 10, 50)
  love.graphics.setColor(255, 255, 255)
  --love.graphics.rectangle("line", self.x, self.y, self.width, self.height)
end



local updateFlames = function(self,dt)
  local keyDown = love.keyboard.isDown
  local left, right, up, down = keyDown("left"), keyDown("right"), keyDown("up"), keyDown("down")
  self.flame.direction = ""
  if up then self.flame.direction = "up" end
  if not self.onGround and not self.gearDown then 
    if left then self.flame.direction = "left"
    elseif right then self.flame.direction = "right"
    elseif down then self.flame.direction = "down" end
  end
  if self.flame.direction ~= "" and not self.remove then self.sound.thrust:play()
  else self.sound.thrust:stop() end
  self.flame.animation:update(dt)
end


local fw, fh = 5, 3
local drawFlames = function(self)
  if self.fuel == 0 then return end
  if self.flame.direction == "down" then self.flame.animation:draw(assets.spritesheet, self.x+self.width/2-self.ox/2, self.y+fh/2-self.oy, math.rad(-90)) end
  if self.flame.direction == "up" then self.flame.animation:draw(assets.spritesheet, self.x+self.width/2, self.y+self.taxiHeight-self.oy-fh/2, math.rad(90)) end
  if self.flame.direction == "left" then self.flame.animation:draw(assets.spritesheet, self.x+self.width+self.ox, self.y+self.height/2+self.oy-fh) end
  if self.flame.direction == "right" then self.flame.animation:draw(assets.spritesheet, self.x-self.ox, self.y+self.height/2+self.oy, math.rad(180)) end
end



local getAssets = function(self)
  self.assets = assets.taxi
  self.sprite = self.assets["Taxi"]
  self.flame  = {direction = "", animation = self.assets["Flame"]:clone()}
  self.sound  = {
    thrust    = assets.getSound("thrust"),
    getIn     = assets.getSound("getIn"),
    getOut    = assets.getSound("getOut"),
    noFuel    = assets.getSound("noFuel"),
    refuel    = assets.getSound("refuel"),
    explosion = assets.getSound("explosion"),
  }
end



local registerSignals = function(self)
  -- Signals
  local passengerDeliveredEvent = function(taxi, passenger)
    if self == taxi then
      self.occupied = false
      self.sound.getOut:play()
      self.sprite = self.assets["Taxi"]
      if passenger then self:addMoney(passenger.money) end
    end
  end
  local passengerEnteredEvent = function(taxi)
    if self == taxi then
      self.occupied = true
      self.sound.getIn:play()
      self.sprite = self.assets["TaxiPassenger"]
    end
  end
  Signal.register("PassengerDelivered", passengerDeliveredEvent)
  Signal.register("PassengerEntered", passengerEnteredEvent)
end



--------------------------------------------------------------------------------------------------------- --
-- Taxi Class --
--------------------------------------------------------------------------------------------------------- --

function Taxi:new(x, y)
  Taxi.super.new(self, x, y)
  for k,v in pairs(default_config) do self[k] = v end
  getAssets(self)
  self:setType("Taxi")
  self:setPriority(2)
  self:setOffset(self.ox, self.oy)
  self:setDimensions(32-self.ox*2, self.taxiHeight-self.oy*2)
  toggleGear(self)
  registerSignals(self)
end



function Taxi:resolveCollision(other)
  if other.type == "TaxiStation" then resolveLandingCollision(self, other)
  elseif other.type == "GasStation" then
    resolveLandingCollision(self, other)
    if self.onGround then self.refueling = true end
  else self:destroy() end
end



function Taxi:destroy()
  self.sound.thrust:stop()
  self.sound.explosion:play()
  if self.occupied then Signal.emit("TaxiExploded", self) end
  Actor.destroy(self)
end



function Taxi:filter(other)
  if other.type == "TaxiStation" or other.type == "GasStation" then return "slide" end
end



function Taxi:addMoney(v)
  self.money = math.min(self.money+v, 1000)
end



function Taxi:removeMoney(v)
  self.money = math.max(self.money-v, 0)
end



function Taxi:update(dt)
  if self.fuel > 0 then updateKeyInput(self, dt) end
  updateBoundaryBehaviour(self, dt)
  updateFuel(self, dt)
  applyMovement(self, dt)
  updateFlames(self, dt)
end



function Taxi:draw()
  drawTaxi(self)
  drawFlames(self)
  drawDebug(self)
end



function Taxi:keypressed(key)
  if key == "up" then Signal.emit("TaxiStarted", self) end
  if key == "space" then toggleGear(self) end
end

return Taxi


